// Copyright (c) 2018 Commissariat à l'énergie atomique et aux énergies alternatives (CEA)
// Copyright (c) 2018 Centre national de la recherche scientifique (CNRS)
// Copyright (c) 2018-2023 Simons Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0.txt
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Authors: Olivier Parcollet, Nils Wentzell

#pragma once
#include <algorithm>
#include <memory>
#include <ranges>
#include <random>
#include "basic_array_view.hpp"

namespace nda {

  // forward for friend declaration
  template <MemoryArray A, typename NewLayoutType>
  auto map_layout_transform(A &&a, NewLayoutType const &new_layout);

  // ---------------------- array--------------------------------

  template <typename ValueType, int Rank, typename LayoutPolicy, char Algebra, typename ContainerPolicy>
  class basic_array {

    static_assert(!std::is_const<ValueType>::value, "ValueType of basic_array cannot be const.");
    static_assert((Algebra != 'N'), " Internal error : Algebra 'N' not supported");
    static_assert((Algebra != 'M') or (Rank == 2), " Internal error : Algebra 'M' only makes sense for rank 2");
    static_assert((Algebra != 'V') or (Rank == 1), " Internal error : Algebra 'V' only makes sense for rank 1");

    public:
    /// Type of the array's values
    using value_type = ValueType;
    /// Type of the memory layout policy
    using layout_policy_t = LayoutPolicy;
    /// Type of the memory layout
    using layout_t = typename LayoutPolicy::template mapping<Rank>;
    /// Type of the container policy
    using container_policy_t = ContainerPolicy;
    /// Type of the memory handle
    using storage_t = typename ContainerPolicy::template handle<ValueType>;
    /// The associated regular type
    using regular_type = basic_array;

    /// The number of dimensions of the array
    static constexpr int rank = Rank;

    static_assert(has_contiguous(layout_t::layout_prop), "Non sense. A basic_array is a contiguous object");

    private:
    // details for the common code with view
    using self_t                   = basic_array;
    using AccessorPolicy           = default_accessor; // no_alias_accessor
    using OwningPolicy             = borrowed<storage_t::address_space>;
    static constexpr bool is_const = false;
    static constexpr bool is_view  = false;

    layout_t lay;
    storage_t sto;

    template <std::integral Int = long>
    basic_array(std::array<Int, Rank> const &shape, mem::init_zero_t) noexcept : lay{shape}, sto{lay.size(), mem::init_zero} {}

    public:
    basic_array_view<ValueType, Rank, LayoutPolicy, 'A', AccessorPolicy, OwningPolicy> as_array_view() { return {*this}; };
    basic_array_view<const ValueType, Rank, LayoutPolicy, 'A', AccessorPolicy, OwningPolicy> as_array_view() const { return {*this}; };

    [[deprecated]] auto transpose()
      requires(Rank == 2)
    {
      return permuted_indices_view<encode(std::array<int, 2>{1, 0})>(*this);
    }
    [[deprecated]] auto transpose() const
      requires(Rank == 2)
    {
      return permuted_indices_view<encode(std::array<int, 2>{1, 0})>(*this);
    }

    // ------------------------------- constructors --------------------------------------------

    /// Empty array
    // Caution! We need to provide a user-defined constructor (over =default)
    // to avoid value initialization of the sso buffer
    basic_array(){};

    ///
    basic_array(basic_array &&X) = default;

    /// Makes a deep copy, since array is a regular type
    explicit basic_array(basic_array const &x) noexcept : lay(x.indexmap()), sto(x.sto) {}

    /// Makes a deep copy, given a basic_array with a different container policy
    template <char AlgebraOther, typename ContainerPolicyOther>
    explicit basic_array(basic_array<ValueType, Rank, LayoutPolicy, AlgebraOther, ContainerPolicyOther> x) noexcept
       : lay(x.indexmap()), sto(std::move(x.storage())) {}

    /** 
     * Construct with a shape [i0, is ...]. 
     * Int are integers (convertible to long), and there must be exactly R arguments.
     * 
     * @param i0, is ... are the extents (lengths) in each dimension
     */
    template <std::integral... Int>
      requires(sizeof...(Int) == Rank)
    explicit basic_array(Int... is) noexcept {
      // Constructing layout and storage in constructor body improves error message for wrong # of args
      lay = layout_t{std::array{long(is)...}};
      sto = storage_t{lay.size()};
    }

    /**
     * Construct one-dimensional array with a shape [i0]
     * with all elements initialized to val
     * Int is an integer (convertible to long)
     *
     * @param i0 is the extents of the only dimension
     */
    template <std::integral Int, typename RHS>
    explicit basic_array(Int i, RHS const &val) noexcept
      requires((Rank == 1 and is_scalar_for_v<RHS, basic_array>))
    {
      lay = layout_t{std::array{long(i)}};
      sto = storage_t{lay.size()};
      assign_from_scalar(val);
    }

    /** 
     * Construct with the given shape and default construct elements
     * 
     * @param shape  Shape of the array (lengths in each dimension)
     */
    template <std::integral Int = long>
    explicit basic_array(std::array<Int, Rank> const &shape) noexcept
      requires(std::is_default_constructible_v<ValueType>)
       : lay(shape), sto(lay.size()) {}

    /// Construct from the layout
    explicit basic_array(layout_t const &layout) noexcept
      requires(std::is_default_constructible_v<ValueType>)
       : lay{layout}, sto{lay.size()} {}

    /// Construct from the layout and existing memory
    explicit basic_array(layout_t const &layout, storage_t &&storage) noexcept : lay{layout}, sto{std::move(storage)} {}

    /** 
     * Constructs from a.shape() and then assign from the evaluation of a
     */
    template <ArrayOfRank<Rank> A>
    basic_array(A const &a) noexcept //
      requires(HasValueTypeConstructibleFrom<A, value_type>)

       : lay(a.shape()), sto{lay.size(), mem::do_not_initialize} {
      static_assert(std::is_constructible_v<value_type, get_value_t<A>>,
                    "Can not construct the array. ValueType can not be constructed from the value_type of the argument");
      if constexpr (std::is_trivial_v<ValueType> or is_complex_v<ValueType>) {
        // simple type. the initialization was not necessary anyway.
        // we use the assign, including the optimization (1d strided, contiguous) possibly
        assign_from_ndarray(a);
      } else {
        // in particular ValueType may or may not be default constructible
        // so we do not init memory, and make the placement new now, directly with the value returned by a
        nda::for_each(lay.lengths(), [&](auto const &...is) { new (sto.data() + lay(is...)) ValueType{a(is...)}; });
      }
    }

    /** 
     * Initialize with any type modelling ArrayInitializer, typically a 
     * delayed operation (mpi operation, matmul) that requires 
     * the knowledge of the data pointer to execute
     *
     */
    template <ArrayInitializer<basic_array> Initializer> // can not be explicit
    basic_array(Initializer const &initializer) noexcept(noexcept(initializer.invoke(basic_array{}))) : basic_array{initializer.shape()} {
      initializer.invoke(*this);
    }

    private: // impl. detail for next function
    static std::array<long, 1> shape_from_init_list(std::initializer_list<ValueType> const &l) noexcept { return {long(l.size())}; }

    template <typename L>
    static auto shape_from_init_list(std::initializer_list<L> const &l) noexcept {
      const auto [min, max] =
         std::minmax_element(std::begin(l), std::end(l), [](auto &&x, auto &&y) { return shape_from_init_list(x) == shape_from_init_list(y); });
      EXPECTS_WITH_MESSAGE(shape_from_init_list(*min) == shape_from_init_list(*max), "initializer list not rectangular !");
      return stdutil::front_append(shape_from_init_list(*max), long(l.size()));
    }

    public:
    ///
    basic_array(std::initializer_list<ValueType> const &l) noexcept //
      requires(Rank == 1)
       : lay(std::array<long, 1>{long(l.size())}), sto{lay.size(), mem::do_not_initialize} {
      long i = 0;
      // We can not assume that ValueType is default constructible. As before, we do not initialize,
      // and use placement new
      // https://godbolt.org/z/Lwic2o. Same code as = for basic type
      // Alternative : if constexpr (std::is_trivial_v<ValueType> or is_complex<ValueType>::value) for (auto const &x : l) *(sto.data() + lay(i++)) = x;
      for (auto const &x : l) { new (sto.data() + lay(i++)) ValueType{x}; }
    }

    ///
    basic_array(std::initializer_list<std::initializer_list<ValueType>> const &l2) noexcept //
      requires(Rank == 2)
       : lay(shape_from_init_list(l2)), sto{lay.size(), mem::do_not_initialize} {
      long i = 0, j = 0;
      for (auto const &l1 : l2) {
        for (auto const &x : l1) { new (sto.data() + lay(i, j++)) ValueType{x}; } // cf dim1
        j = 0;
        ++i;
      }
    }

    ///
    basic_array(std::initializer_list<std::initializer_list<std::initializer_list<ValueType>>> const &l3) noexcept //
      requires(Rank == 3)
       : lay(shape_from_init_list(l3)), sto{lay.size(), mem::do_not_initialize} {
      long i = 0, j = 0, k = 0;
      static_assert(Rank == 3, "?");
      for (auto const &l2 : l3) {
        for (auto const &l1 : l2) {
          for (auto const &x : l1) { new (sto.data() + lay(i, j, k++)) ValueType{x}; } // cf dim1
          k = 0;
          ++j;
        }
        j = 0;
        ++i;
      }
    }

    /////
    //template <typename U>
    //explicit basic_array(std::initializer_list<std::initializer_list<U>> const &l2) noexcept //
    //requires((Rank == 1) and (std::is_constructible_v<ValueType, std::initializer_list<U>>))
    //: lay(l2.size()), sto{lay.size(), mem::do_not_initialize} {
    //long i = 0;
    //for (auto const &l1 : l2) { new (sto.data() + lay(i++)) ValueType{l1}; }
    //}

    /////
    //template <typename U>
    //explicit basic_array(std::initializer_list<std::initializer_list<std::initializer_list<U>>> const &l3) noexcept //
    //requires((Rank == 2) and (std::is_constructible_v<ValueType, std::initializer_list<U>>))
    //: lay(pop(shape_from_init_list(l3))), sto{lay.size(), mem::do_not_initialize} {
    //long i = 0, j = 0;
    //static_assert(Rank == 3, "?");
    //for (auto const &l2 : l3) {
    //for (auto const &l1 : l2) { new (sto.data() + lay(i, j++)) ValueType{l1}; }
    //j = 0;
    //++i;
    //}
    /*}*/

    /// Allows to move a array of Rank 2 into a matrix and vice versa
    /// Beware that for stack/sso array, it will copy the data (but move them for heap allocated array).
    /// \trailing_requires
    template <char Algebra2>
    explicit basic_array(basic_array<ValueType, 2, LayoutPolicy, Algebra2, ContainerPolicy> &&am) noexcept
      requires(Rank == 2) // NB Rank =2 since matrix/array for the moment. generalize if needed
       : basic_array{am.indexmap(), std::move(am).storage()} {}

    //------------------ Factories -------------------------

    /// Make a zero-initialized array of the given shape
    template <std::integral Int = long>
    static basic_array zeros(std::array<Int, Rank> const &shape)
      requires(std::is_standard_layout_v<ValueType> && std::is_trivially_copyable_v<ValueType>)
    {
      return basic_array{stdutil::make_std_array<long>(shape), mem::init_zero};
    }

    /// Make a zero-initialized array with the given dimensions
    template <std::integral... Ints>
    static basic_array zeros(Ints... i)
      requires(sizeof...(Ints) == Rank)
    {
      return zeros(std::array<long, Rank>{i...});
    }

    /// Make an array of the given shape holding 'scalar ones'
    template <std::integral Int = long>
    static basic_array ones(std::array<Int, Rank> const &shape)
      requires(nda::is_scalar_v<ValueType>)
    {
      auto res = basic_array{stdutil::make_std_array<long>(shape)};
      res()    = ValueType{1};
      return res;
    }

    /// Make an array of the given dimensions holding 'scalar ones'
    template <std::integral... Ints>
    static basic_array ones(Ints... i)
      requires(sizeof...(Ints) == Rank)
    {
      return ones(std::array<long, Rank>{i...});
    }

    /// Create an array the given shape and populate it with random
    /// samples from a uniform distribution over [0, 1)
    template <std::integral Int = long>
    static basic_array rand(std::array<Int, Rank> const &shape)
      requires(std::is_floating_point_v<ValueType> or nda::is_complex_v<ValueType>)
    {

      auto static gen  = std::mt19937(std::random_device{}());
      auto static dist = std::uniform_real_distribution<>(0.0, 1.0);

      auto res = basic_array{shape};
      if constexpr (nda::is_complex_v<ValueType>)
        for (auto &x : res) x = dist(gen) + 1i * dist(gen);
      else
        for (auto &x : res) x = dist(gen);

      return res;
    }

    /// Create an array the given dimensions and populate it with random
    /// samples from a uniform distribution over [0, 1)
    template <std::integral... Ints>
    static basic_array rand(Ints... i)
      requires(sizeof...(Ints) == Rank)
    {
      return rand(std::array<long, Rank>{i...});
    }

    //------------------ Assignment -------------------------

    ///
    basic_array &operator=(basic_array &&x) = default;

    /// Deep copy (array is a regular type). Invalidates all references to the storage.
    basic_array &operator=(basic_array const &X) = default;

    /// Deep copy assignment given array with different algebra and/or container policy
    template <char Algebra_other, typename ContainerPolicy_other>
    basic_array &operator=(basic_array<ValueType, Rank, LayoutPolicy, Algebra_other, ContainerPolicy_other> const &x) {
      *this = basic_array{x};
      return *this;
    }

    /** 
     * Resizes the array (if necessary).
     * Invalidates all references to the storage.
     *
     * @tparam RHS A scalar or an object modeling NdArray
     */
    template <ArrayOfRank<Rank> RHS>
    basic_array &operator=(RHS const &rhs) noexcept {
      static_assert(!is_const, "Cannot assign to a const !");
      resize(rhs.shape());
      assign_from_ndarray(rhs); // common code with view, private
      return *this;
    }

    /** 
     * Resizes the array (if necessary).
     * Invalidates all references to the storage.
     *
     * @tparam RHS A scalar or an object modeling NdArray
     */
    template <typename RHS>
    // FIXME : explode this notion
    basic_array &operator=(RHS const &rhs) noexcept
      requires(is_scalar_for_v<RHS, basic_array>)
    {
      static_assert(!is_const, "Cannot assign to a const !");
      assign_from_scalar(rhs); // common code with view, private
      return *this;
    }

    /** 
     * 
     */
    template <ArrayInitializer<basic_array> Initializer>
    basic_array &operator=(Initializer const &initializer) noexcept {
      resize(initializer.shape());
      initializer.invoke(*this);
      return *this;
    }

    //------------------ resize  -------------------------
    /** 
     * Resizes the array.
     * Invalidates all references to the storage.
     * Content is undefined, makes no copy of previous data.
     *
     */
    template <std::integral... Int>
    void resize(Int const &...extent) {
      static_assert(std::is_copy_constructible_v<ValueType>, "Can not resize an array if its value_type is not copy constructible");
      static_assert(sizeof...(extent) == Rank, "Incorrect number of arguments for resize. Should be Rank");
      resize(std::array<long, Rank>{long(extent)...});
    }

    /** 
     * Resizes the array.
     * Invalidates all references to the storage.
     * Content is undefined, makes no copy of previous data.
     *
     * @param shape  New shape of the array (lengths in each dimension)
     */
    [[gnu::noinline]] void resize(std::array<long, Rank> const &shape) {
      lay = layout_t(shape);
      // Construct a storage only if the new index is not compatible (size mismatch).
      if (sto.is_null() or (sto.size() != lay.size())) sto = storage_t{lay.size()};
    }

#include "./_impl_basic_array_view_common.hpp"
  };

  // --- Class Template Argument Deduction Guides ---

  template <MemoryArray A>
  basic_array(A &&a) -> basic_array<get_value_t<A>, get_rank<A>, get_contiguous_layout_policy<get_rank<A>, get_layout_info<A>.stride_order>,
                                    get_algebra<A>, heap<mem::get_addr_space<A>>>;

  template <Array A>
  basic_array(A &&a) -> basic_array<get_value_t<A>, get_rank<A>, C_layout, get_algebra<A>, heap<>>;

  // --- get_regular_t ---

  template <typename T, typename T2 = std::remove_reference_t<T> /* Keep this: Fix for gcc11 bug */>
  using get_regular_t = decltype(basic_array{std::declval<T>()});

  // --- Get the associated regular type with host/device/unified memory ---

  template <typename T, typename RT = get_regular_t<T>>
  using get_regular_host_t =
     std::conditional_t<mem::on_host<RT>, RT,
                        basic_array<get_value_t<RT>, get_rank<RT>, get_contiguous_layout_policy<get_rank<RT>, get_layout_info<RT>.stride_order>,
                                    get_algebra<RT>, heap<mem::Host>>>;

  template <typename T, typename RT = get_regular_t<T>>
  using get_regular_device_t =
     std::conditional_t<mem::on_device<RT>, RT,
                        basic_array<get_value_t<RT>, get_rank<RT>, get_contiguous_layout_policy<get_rank<RT>, get_layout_info<RT>.stride_order>,
                                    get_algebra<RT>, heap<mem::Device>>>;

  template <typename T, typename RT = get_regular_t<T>>
  using get_regular_unified_t =
     std::conditional_t<mem::on_unified<RT>, RT,
                        basic_array<get_value_t<RT>, get_rank<RT>, get_contiguous_layout_policy<get_rank<RT>, get_layout_info<RT>.stride_order>,
                                    get_algebra<RT>, heap<mem::Unified>>>;

} // namespace nda

#include "./layout_transforms.hpp"
